from typing import Any, Dict, Optional

from pydantic import BaseModel

from combadge.support.http.abc import SupportsBody
from combadge.support.rest.abc import RequiresMethod, RequiresPath, SupportsQueryParams


class Request(RequiresMethod, RequiresPath, SupportsBody, SupportsQueryParams, BaseModel):
    """Backend-agnostic REST request."""

    def body_dict(self) -> Optional[Dict[str, Any]]:
        """Convert the body to a dictionary."""
        return body.dict(by_alias=True) if (body := self.body) is not None else None



"""Generic marks applicable to a variety of protocols."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable, Dict, Tuple, TypeVar, Union

from typing_extensions import TypeAlias

from combadge.core.mark import MethodMark, ParameterMark
from combadge.core.typevars import Identity
from combadge.support.rest.abc import RequiresMethod, RequiresPath, SupportsQueryParams

T = TypeVar("T")


class _PathMark(MethodMark[RequiresPath]):
    _factory: Callable[..., str]
    __slots__ = ("_factory",)

    def __init__(self, path_or_factory: Union[str, Callable[..., str]]) -> None:  # noqa: D107
        if callable(path_or_factory):
            self._factory = path_or_factory
        else:
            self._factory = path_or_factory.format

    def prepare_request(  # noqa: D102
        self,
        request: RequiresPath,
        args: Tuple[Any, ...],
        kwargs: Dict[str, Any],
    ) -> None:
        request.path = self._factory(*args, **kwargs)


def path(path_or_factory: Union[str, Callable[..., str]]) -> Identity:
    """
    Specify a URL path.

    Example:
        >>> @path("/hello/world")
        >>> def call() -> None: ...

        >>> @path("/hello/{name}")
        >>> def call(name: str) -> None: ...

        >>> @path(lambda name, **_: f"/hello/{name}")
        >>> def call(name: str) -> None: ...
    """
    return _PathMark(path_or_factory).mark


@dataclass
class _RestMethodMark(MethodMark[RequiresMethod]):
    method: str  # TODO: enum?

    def prepare_request(  # noqa: D102
        self,
        request: RequiresMethod,
        _args: Tuple[Any, ...],
        _kwargs: Dict[str, Any],
    ) -> None:
        request.method = self.method


def method(method: str) -> Identity:
    """Specify a HTTP/REST method."""
    return _RestMethodMark(method).mark


@dataclass
class QueryParameterMark(ParameterMark[SupportsQueryParams]):
    """Designates parameter as a service call's query parameter."""

    name: str
    __slots__ = ("name",)

    def prepare_request(self, request: SupportsQueryParams, value: Any) -> None:  # noqa: D102
        request.query_params.append((self.name, value))


QueryParam: TypeAlias = QueryParameterMark



"""Generic response models."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Type, Union

from pydantic import BaseModel
from typing_extensions import NoReturn, Self


class BaseResponse(ABC, BaseModel):
    """Base model representing any possible service response."""

    @abstractmethod
    def raise_for_result(self) -> Union[None, NoReturn]:
        """Raise an exception if the service call has failed."""
        raise NotImplementedError

    @abstractmethod
    def expect(self, exception_type: Type[BaseException], *args: Any) -> Union[None, NoReturn]:
        """Raise the specified exception if the service call has failed."""
        raise NotImplementedError

    @abstractmethod
    def unwrap(self) -> Union[Self, NoReturn]:
        """Return itself if the call was successful, raises an exception otherwise."""
        raise NotImplementedError


class SuccessfulResponse(BaseResponse):
    """
    Parent model for successful responses.

    Note:
        - May be used directly when no payload is expected.
    """

    def raise_for_result(self) -> None:
        """
        Do nothing.

        This call is a no-op since the response is successful.
        """

    def expect(self, _exception_type: Type[BaseException], *_args: Any) -> Union[None, NoReturn]:
        """Do nothing."""

    def unwrap(self) -> Self:
        """Return itself since there's no error."""
        return self


class FaultyResponse(BaseResponse, ABC):
    """
    Parent model for faulty responses (errors).

    Notes:
        - Useful when server returns errors in a free from (for example, an `<error>` tag).
        - Must be subclassed.
        - For SOAP Fault use or subclass the specialized `GenericSoapFault`.
    """

    class Error(Exception):
        """
        Dynamically derived exception class.

        The problem with `pydantic` is that you can't inherit from `BaseModel` and `Exception`
        at the same time. The workaround is to dynamically construct a derived exception class
        that is available via the class attribute and raised by `raise_for_result()`.
        """

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Build the derived exception class."""

        error_bases = tuple(base.Error for base in cls.__bases__ if issubclass(base, FaultyResponse))

        class DerivedException(*error_bases):  # type: ignore
            """
            Derived exception class.

            Notes:
                - This docstring is overridden by the corresponding model docstring.
            """

        DerivedException.__name__ = f"{cls.__name__}.Error"
        DerivedException.__qualname__ = f"{cls.__qualname__}.Error"
        DerivedException.__doc__ = cls.__doc__ or DerivedException.__doc__
        cls.Error = DerivedException  # type: ignore

    def raise_for_result(self) -> NoReturn:
        """Raise the derived exception."""
        raise self.Error

    def expect(self, exception_type: Type[BaseException], *args: Any) -> Union[None, NoReturn]:
        """Raise the specified exception with the derived exception context."""
        raise exception_type(*args) from self.Error

    def unwrap(self) -> NoReturn:
        """Raise the derived exception."""
        raise self.Error



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, Generic, List, Tuple, Type, TypeVar

from typing_extensions import Annotated, get_origin
from typing_extensions import get_args as get_type_args

from combadge.core.typevars import Identity, RequestT

T = TypeVar("T")


class MethodMark(ABC, Generic[RequestT]):
    """Method-specific mark that modifies an entire request based on all the call arguments."""

    __slots__ = ()

    def wrap(self, what: T) -> T:
        """
        Wrap the argument according to the mark.

        Notes:
            - Does nothing by default. Override in a child class.
        """
        return what

    def prepare_request(self, request: RequestT, args: Tuple[Any, ...], kwargs: Dict[str, Any]) -> None:
        """
        Modify the request according to the mark.

        Notes:
            - Does nothing by default. Override in a child class.

        Args:
            request: request that is being constructed, please refer to the ABCs for relevant keys
            args: service call positional arguments
            kwargs: service call keyword arguments
        """

    @staticmethod
    def ensure_marks(in_: Any) -> List[MethodMark]:
        """Ensure that the argument contains the mark list attribute, and return the list."""
        try:
            marks = in_.__combadge_marks__
        except AttributeError:
            marks = in_.__combadge_marks__ = []
        return marks

    def mark(self, what: T) -> T:
        """
        Mark the argument with itself.

        This is not a part of the public interface and is used to derive the decorators.
        """
        MethodMark.ensure_marks(what).append(self)
        return what


class ParameterMark(Generic[RequestT], ABC):
    """Parameter-specific mark that modifies a request with a call-time argument."""

    __slots__ = ()

    @staticmethod
    def extract(type_: Type[Any]) -> List[ParameterMark]:
        """Extract all parameter marks from the type annotation."""
        if get_origin(type_) is Annotated:
            return [arg for arg in get_type_args(type_) if isinstance(arg, ParameterMark)]
        return []

    @abstractmethod
    def prepare_request(self, request: RequestT, value: Any) -> None:
        """Update the request according to the mark and the actual argument."""


class _DecorateMethodMark(MethodMark[Any]):
    _decorator: Identity
    __slots__ = ("_decorator",)

    def __init__(self, decorator: Identity) -> None:
        self._decorator = decorator

    def wrap(self, what: T) -> T:
        return self._decorator(what)


def decorator(decorator: Identity[T]) -> Identity[T]:
    """
    Put the decorator on top of the generated bound service method.

    Example:
        >>> @decorator(functools.cache())
        >>> def service_method(self, ...) -> ...:
        >>>     ...
    """
    return _DecorateMethodMark(decorator).mark



from typing import Any, Type

from pydantic import BaseModel
from zeep.exceptions import Fault
from zeep.proxy import AsyncOperationProxy, AsyncServiceProxy

from combadge.core.binder import BaseBoundService, Signature
from combadge.core.interfaces import SupportsBindServiceMethod, SupportsServiceMethodCall
from combadge.core.request import build_request
from combadge.core.typevars import ResponseT
from combadge.support.soap.request import Request
from combadge.support.soap.response import SoapFaultT
from combadge.support.zeep.backends.base import BaseZeepBackend


class ZeepBackend(BaseZeepBackend[AsyncServiceProxy, AsyncOperationProxy], SupportsBindServiceMethod):
    """Asynchronous Zeep service."""

    async def __call__(
        self,
        request: Request,
        response_type: Type[ResponseT],
        fault_type: Type[SoapFaultT],
    ) -> BaseModel:
        """
        Call the specified service method.

        Args:
            request: prepared request
            response_type: non-fault response model type
            fault_type: SOAP fault model type
        """
        operation = self._get_operation(request.operation_name)
        try:
            response = await operation(**request.body.dict(by_alias=True))
        except Fault as e:
            return self._parse_soap_fault(e, fault_type)
        return self._parse_response(response, response_type)

    def bind_method(self, signature: Signature) -> SupportsServiceMethodCall:  # noqa: D102
        response_type, fault_type = self._split_response_type(signature.return_type)

        async def resolved_method(service: BaseBoundService, *args: Any, **kwargs: Any) -> BaseModel:
            request = build_request(Request, signature, service, args, kwargs)
            return await self(request, response_type, fault_type)

        return resolved_method  # type: ignore[return-value]



from __future__ import annotations

from abc import abstractmethod
from typing import Any

from pydantic import BaseModel
from typing_extensions import Protocol, Self

from combadge.core.binder import BaseBoundService, Signature, bind


class SupportsService(Protocol):
    """
    Convenience protocol that forwards the `bind` call.

    You can still inherit from `Protocol` directly and call `bind` manually.
    """

    @classmethod
    def bind(cls, to_backend: SupportsBindServiceMethod) -> Self:
        """Bind the protocol to the specified backend."""
        return bind(cls, to_backend)


class SupportsBindServiceMethod(Protocol):
    """Supports binding a method to the current instance."""

    @abstractmethod
    def bind_method(self, signature: Signature) -> SupportsServiceMethodCall:
        """
        «Binds» the `method` to the current instance (for example, a backend).

        Args:
            signature: extracted method signature

        Returns:
            Callable service method which is fully capable of sending a request and receiving a response.
        """
        raise NotImplementedError


class SupportsServiceMethodCall(Protocol):
    """
    Bound method call specification.

    Usually implemented by a backend in its `bind_method`.
    """

    @abstractmethod
    def __call__(self, __service: BaseBoundService, *__args: Any, **__kwargs: Any) -> BaseModel:
        """
        Call the service method.

        Args:
            __service: bound service instance
            __args: positional request parameters
            __kwargs: keyword request parameters

        Returns:
            Parsed response model.
        """
        raise NotImplementedError



from abc import ABC
from typing import Any, Generic, Tuple, Type, TypeVar, Union

from pydantic import BaseModel, parse_obj_as
from typing_extensions import get_args as get_type_args
from typing_extensions import get_origin as get_type_origin
from zeep.exceptions import Fault
from zeep.helpers import serialize_object
from zeep.proxy import OperationProxy, ServiceProxy

from combadge.core.response import SuccessfulResponse
from combadge.core.typevars import ResponseT
from combadge.support.soap.response import BaseSoapFault, SoapFaultT

ServiceProxyT = TypeVar("ServiceProxyT", bound=ServiceProxy)
"""
Specific service proxy type returned by a Zeep client.

See Also:
    - https://docs.python-zeep.org/en/master/client.html#the-serviceproxy-object
"""

OperationProxyT = TypeVar("OperationProxyT", bound=OperationProxy)


class BaseZeepBackend(ABC, Generic[ServiceProxyT, OperationProxyT]):
    """Base class for the sync and async backends. Not intended for a direct use."""

    __slots__ = ("_service",)

    def __init__(self, service: ServiceProxyT) -> None:
        """Instantiate the backend."""
        self._service = service

    @staticmethod
    def _split_response_type(response_type: Type[Any]) -> Tuple[Type[BaseModel], Type[BaseSoapFault]]:
        """
        Split the response type into non-faults and faults.

        SOAP faults are handled separately, so we need to extract them from the annotated
        response type.
        """

        if get_type_origin(response_type) is Union:  # noqa: SIM108
            return_types = get_type_args(response_type)
        else:
            return_types = (response_type,)

        fault_type = BaseSoapFault

        response_type: Any = ...
        for return_type in return_types:
            if isinstance(return_type, type) and issubclass(return_type, BaseSoapFault):
                fault_type = Union[return_type, fault_type]  # type: ignore
            elif response_type is Ellipsis:
                response_type = return_type
            else:
                response_type = Union[response_type, return_type]  # type: ignore

        if response_type is Ellipsis:
            response_type = SuccessfulResponse

        return response_type, fault_type

    def _get_operation(self, name: str) -> OperationProxyT:
        """Get an operation by its name."""
        try:
            return self._service[name]
        except AttributeError as e:
            raise RuntimeError(f"available operations are: {dir(self._service)}") from e

    @staticmethod
    def _parse_response(value: Any, response_type: Type[ResponseT]) -> ResponseT:
        """Parse the response value using the generic response types."""
        return parse_obj_as(response_type, serialize_object(value, dict))

    @staticmethod
    def _parse_soap_fault(exception: Fault, fault_type: Type[SoapFaultT]) -> SoapFaultT:
        """Parse the SOAP fault."""
        return parse_obj_as(fault_type, exception.__dict__)



from typing import Any, Iterable, Mapping, Type, Union

from pydantic import validate_model
from typing_extensions import NoReturn

from combadge.core.binder import BaseBoundService, Signature
from combadge.core.typevars import RequestT


def build_request(
    request_class: Type[RequestT],
    signature: Signature,
    service: BaseBoundService,
    call_args: Iterable[Any],
    call_kwargs: Mapping[str, Any],
) -> Union[RequestT, NoReturn]:
    """
    Build a request using the provided request type, marks, and service call arguments.

    Args:
        request_class: target back-end request class
        signature: extracted description of the protocol service method
        service: an instance of a service class, on which the method is being called
        call_args: service method call positional arguments
        call_kwargs: service method call keyword arguments
    """

    bound_arguments = signature.bind_arguments(service, *call_args, **call_kwargs)
    bound_arguments.apply_defaults()

    all_arguments = bound_arguments.arguments
    call_args = bound_arguments.args
    call_kwargs = bound_arguments.kwargs

    # Construct an initial empty request without validation.
    # See also: https://github.com/pydantic/pydantic/issues/1864#issuecomment-679044432
    request = request_class.construct()

    # Apply the method marks: they receive all the arguments at once.
    for mark in signature.method_marks:
        mark.prepare_request(request, call_args, call_kwargs)

    # Apply the parameter marks: they receive their respective values.
    for name, mark in signature.parameter_marks:
        try:
            value = all_arguments[name]
        except KeyError:
            pass
        else:
            mark.prepare_request(request, value)

    # Validate and return the request.
    *_, error = validate_model(request_class, request.__dict__)
    if error:
        raise error
    return request



"""Generic HTTP headers."""

from typing import TypeVar

from typing_extensions import Annotated, TypeAlias

from combadge.support.http.marks import Header

T = TypeVar("T")

AcceptLanguage: TypeAlias = Annotated[T, Header("Accept-Language")]
# TODO: more headers.
# TODO: typed headers?



"""Mixins for concrete RESTful API request classes."""

from abc import ABC
from typing import Any, List, Tuple

from pydantic import BaseModel, Field
from typing_extensions import Annotated


class RequiresPath(ABC, BaseModel):
    """Requires a request URL path."""

    path: str


class RequiresMethod(ABC, BaseModel):
    """Requires a request method (for example, HTTP method)."""

    method: str


class SupportsQueryParams(ABC, BaseModel):
    """Supports query parameters."""

    query_params: Annotated[List[Tuple[str, Any]], Field(default_factory=list)]



from __future__ import annotations

from dataclasses import dataclass
from functools import update_wrapper
from inspect import BoundArguments
from inspect import getmembers as get_members
from inspect import signature as get_signature
from typing import TYPE_CHECKING, Any, Callable, Iterable, List, Mapping, Tuple, Type, TypeVar

from typing_extensions import ParamSpec

try:
    from inspect import get_annotations  # type: ignore[attr-defined]
except ImportError:
    from get_annotations import get_annotations  # type: ignore[no-redef]

from pydantic import BaseModel

from combadge.core.mark import MethodMark, ParameterMark
from combadge.core.response import SuccessfulResponse

if TYPE_CHECKING:
    from combadge.core.interfaces import SupportsBindServiceMethod, SupportsServiceMethodCall
    from combadge.core.typevars import ServiceProtocolT

T = TypeVar("T")
P = ParamSpec("P")


class BaseBoundService:
    """Parent of all bound service instances."""


def bind(from_protocol: Type["ServiceProtocolT"], to_backend: SupportsBindServiceMethod) -> "ServiceProtocolT":
    """
    Hereinafter «binding» is constructing a callable service instance from the protocol specification.

    This function returns an instance which implements the specified protocol
    by calling the specified backend.

    Args:
        from_protocol: service protocol description, used to extract request and response types etc.
        to_backend: backend which should perform the service requests
    """

    class BoundService(BaseBoundService, from_protocol):  # type: ignore[misc, valid-type]
        """Bound service class that implements the protocol."""

    for name, method in _enumerate_methods(from_protocol):
        signature = Signature.from_method(method)
        resolved_method: SupportsServiceMethodCall = to_backend.bind_method(signature)
        resolved_method = _wrap(resolved_method, signature.method_marks)
        update_wrapper(resolved_method, method)
        setattr(BoundService, name, resolved_method)

    _update_bound_service(BoundService, from_protocol)
    return BoundService()


def _wrap(method: Callable[P, T], with_marks: Iterable[MethodMark]) -> Callable[P, T]:
    for mark in with_marks:
        method = mark.wrap(method)
    return method


def _enumerate_methods(of_protocol: type) -> Iterable[tuple[str, Any]]:
    """Enumerate the service protocol methods."""

    for name, method in get_members(of_protocol, callable):
        if name.startswith("_"):
            continue
        parameters = get_signature(method).parameters
        if "self" not in parameters:
            continue
        yield name, method


def _update_bound_service(service_class: Type[BaseBoundService], with_protocol: Type[Any]) -> None:
    """Update the generated service class' magic attributes."""

    del service_class.__abstractmethods__  # type: ignore[attr-defined]
    service_class.__name__ = f"{service_class.__name__}[{with_protocol.__name__}]"
    service_class.__qualname__ = f"{service_class.__qualname__}[{with_protocol.__qualname__}]"
    service_class.__doc__ = service_class.__doc__


@dataclass
class Signature:
    """
    Contains extracted information about a service method.

    Why? Because passing all these parameters into `bind_method` would be messy.
    """

    bind_arguments: Callable[..., BoundArguments]
    method_marks: List[MethodMark]
    parameter_marks: List[Tuple[str, ParameterMark]]
    return_type: Type[BaseModel]

    @classmethod
    def from_method(cls, method: Any) -> Signature:
        """Create a signature from the specified method."""
        type_hints = get_annotations(method, eval_str=True)
        return Signature(
            bind_arguments=get_signature(method).bind,
            method_marks=MethodMark.ensure_marks(method),
            parameter_marks=list(cls._extract_parameter_marks(type_hints)),
            return_type=cls._extract_return_type(type_hints),
        )

    @staticmethod
    def _extract_parameter_marks(annotations: Mapping[str, Any]) -> Iterable[Tuple[str, ParameterMark]]:
        """Extract all parameter marks for all the parameters."""
        for name, hint in annotations.items():
            for mark in ParameterMark.extract(hint):
                yield name, mark

    @staticmethod
    def _extract_return_type(annotations: Mapping[str, Any]) -> Type[BaseModel]:
        """Extract return type from the method."""
        return annotations.get("return", SuccessfulResponse)



"""Marks for HTTP-compatible requests."""

from dataclasses import dataclass
from typing import Any, TypeVar, Union

from typing_extensions import Annotated, TypeAlias

from combadge.core.mark import ParameterMark
from combadge.support.http.abc import RequiresBody, SupportsBody, SupportsHeaders

T = TypeVar("T")


@dataclass
class BodyParameterMark(ParameterMark[Union[RequiresBody, SupportsBody]]):
    """Designates a parameter a service call's request body."""

    __slots__ = ()

    def prepare_request(self, request: Union[RequiresBody, SupportsBody], value: Any) -> None:  # noqa: D102
        request.body = value


Body: TypeAlias = Annotated[T, BodyParameterMark()]


@dataclass
class HeaderParameterMark(ParameterMark[SupportsHeaders]):
    """Designates parameter as a service call's additional header."""

    name: str
    __slots__ = ("name",)

    def prepare_request(self, request: SupportsHeaders, value: Any) -> None:  # noqa: D102
        request.headers.append((self.name, value))


Header: TypeAlias = HeaderParameterMark



"""Response extensions for SOAP."""

from typing import NoReturn, TypeVar

from combadge.core.response import FaultyResponse


class BaseSoapFault(FaultyResponse):
    """
    SOAP Fault model.

    Notes:
        - This class matches the SOAP Fault specification.
          For custom errors returned in a SOAP response body (such as `<error>` tag),
          subclass the `FaultyResponse`.

    See Also:
        - https://www.w3.org/TR/2000/NOTE-SOAP-20000508/#_Toc478383507
    """

    code: str
    message: str

    def raise_for_result(self) -> NoReturn:
        """Raise itself always."""
        raise self.Error(f"{self.code}: {self.message}")


SoapFaultT = TypeVar("SoapFaultT", bound=BaseSoapFault)
"""Specific SOAP Fault model type."""



"""Mixins for SOAP request classes."""

from dataclasses import dataclass
from typing import Any, Dict, Tuple

from combadge.core.mark import MethodMark
from combadge.core.typevars import Identity
from combadge.support.soap.abc import RequiresOperationName


@dataclass
class _OperationNameMethodMark(MethodMark[RequiresOperationName]):
    name: str

    __slots__ = ("name",)

    def prepare_request(  # noqa: D102
        self,
        request: RequiresOperationName,
        _args: Tuple[Any, ...],
        _kwargs: Dict[str, Any],
    ) -> None:
        request.operation_name = self.name


def operation_name(name: str) -> Identity:
    """Designates a service call's operation name (for example, SOAP name)."""
    return _OperationNameMethodMark(name).mark



"""Mixins for HTTP-related request classes."""

from abc import ABC
from typing import Any, List, Optional, Tuple

from pydantic import BaseModel, Field
from typing_extensions import Annotated


class SupportsHeaders(ABC, BaseModel):
    """Adds support for additional HTTP headers."""

    headers: Annotated[List[Tuple[str, Any]], Field(default_factory=list)]


class RequiresBody(ABC, BaseModel):
    """Requires a request body (for example, a JSON payload or a SOAP body)."""

    body: BaseModel


class SupportsBody(ABC, BaseModel):
    """Supports an optional request body (for example, a JSON payload or a SOAP body)."""

    body: Optional[BaseModel] = None



from typing import Any, Type

from httpx import Client, Response
from pydantic import BaseModel, parse_obj_as

from combadge.core.binder import BaseBoundService, Signature
from combadge.core.interfaces import SupportsBindServiceMethod, SupportsServiceMethodCall
from combadge.core.request import build_request
from combadge.core.typevars import ResponseT
from combadge.support.rest.request import Request


class HttpxBackend(SupportsBindServiceMethod):
    """
    Sync HTTPX backend for REST APIs.

    See Also:
        - https://www.python-httpx.org/
    """

    _client: Client
    __slots__ = ("_client",)

    def __init__(self, client: Client) -> None:  # noqa: D107
        self._client = client

    def __call__(self, request: Request, response_type: Type[ResponseT]) -> ResponseT:
        """Call the backend."""
        response: Response = self._client.request(
            request.method,
            request.path,
            json=request.body_dict(),
            params=request.query_params,
        )
        response.raise_for_status()
        return parse_obj_as(response_type, response.json())

    def bind_method(self, signature: Signature) -> SupportsServiceMethodCall:  # noqa: D102
        def resolved_method(service: BaseBoundService, *args: Any, **kwargs: Any) -> BaseModel:
            request = build_request(Request, signature, service, args, kwargs)
            return self(request, signature.return_type)

        return resolved_method  # type: ignore[return-value]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['combadge',
 'combadge.core',
 'combadge.support',
 'combadge.support.http',
 'combadge.support.httpx',
 'combadge.support.httpx.backends',
 'combadge.support.rest',
 'combadge.support.soap',
 'combadge.support.zeep',
 'combadge.support.zeep.backends']

package_data = \
{'': ['*']}

install_requires = \
['pydantic>=1.10.4,<2.0.0', 'typing-extensions>=4.4.0,<5.0.0']

extras_require = \
{':python_version < "3.10"': ['get-annotations>=0.1.2,<0.2.0'],
 'httpx': ['httpx>=0.23.3,<0.24.0'],
 'zeep': ['zeep>=4.2.1,<5.0.0']}

setup_kwargs = {
    'name': 'combadge',
    'version': '0.1.0.dev2',
    'description': 'Generic API client based on Pydantic',
    'long_description': '# `combadge`\n\n> 📻 Application to the service, please respond!\n\n[![Checks](https://img.shields.io/github/checks-status/kpn/combadge/main)](https://github.com/kpn/combadge/actions/workflows/check.yaml)\n[![Coverage](https://codecov.io/gh/kpn/combadge/branch/main/graph/badge.svg?token=ZAqYAaTXwE)](https://codecov.io/gh/kpn/combadge)\n![Code style](https://img.shields.io/badge/code%20style-black-000000.svg)\n[![License](https://img.shields.io/github/license/kpn/combadge)](LICENSE)\n\n## Features\n\n- [**Pydantic**](https://docs.pydantic.dev/)-based request and response models\n- Automatically derived exception classes\n- Using [**Protocol**](https://peps.python.org/pep-0544/)s to define service classes\n- Built-in backends:\n  - [HTTPX](https://www.python-httpx.org/), sync and async\n  - [Zeep](https://docs.python-zeep.org/en/master/), sync and async\n- Pluggable backends\n\n## 🚀 Quickstart\n\nℹ️ This `README` is [tested](tests/integration/test_readme.py) and should run «as is».\n\n### 🦋 [HTTPX](https://www.python-httpx.org/) backend\n\n```python\n# test_id=quickstart_httpx\n\nfrom typing import List\n\nfrom httpx import Client\nfrom pydantic import BaseModel, Field\nfrom typing_extensions import Annotated, Protocol\n\nfrom combadge.core.binder import bind\nfrom combadge.support.httpx.backends.sync import HttpxBackend\nfrom combadge.support.rest.marks import QueryParam, method, path\n\n\n# 1️⃣ Declare the response models:\nclass CurrentCondition(BaseModel):\n    humidity: int\n    temperature: Annotated[float, Field(alias="temp_C")]\n\n\nclass Weather(BaseModel):\n    current: Annotated[List[CurrentCondition], Field(alias="current_condition")]\n\n\n# 2️⃣ Declare the protocol:\nclass SupportsWttrIn(Protocol):\n    @method("GET")\n    @path("/{in_}")\n    def get_weather(\n        self,\n        *,\n        in_: str,\n        format_: Annotated[str, QueryParam("format")] = "j1",\n    ) -> Weather:\n        raise NotImplementedError\n\n\n# 3️⃣ Bind the service:\nbackend = HttpxBackend(Client(base_url="https://wttr.in"))\nservice = bind(SupportsWttrIn, backend)\n\n# 🚀 Call the service:\nresponse = service.get_weather(in_="amsterdam")\nassert response.current[0].humidity == 71\nassert response.current[0].temperature == 8.0\n```\n\n### 🧼 [Zeep](https://docs.python-zeep.org/en/master/) backend\n\n```python\n# test_id=quickstart_zeep\n\nfrom typing import Literal, Union\n\nimport zeep\nfrom pydantic import BaseModel, Field\nfrom pytest import raises\nfrom typing_extensions import Annotated\n\nfrom combadge.core.interfaces import SupportsService\nfrom combadge.core.response import FaultyResponse, SuccessfulResponse\nfrom combadge.support.http.marks import Body\nfrom combadge.support.soap.marks import operation_name\nfrom combadge.support.zeep.backends.sync import ZeepBackend\n\n\n# 1️⃣ Declare the request model:\nclass NumberToWordsRequest(BaseModel, allow_population_by_field_name=True):\n    number: Annotated[int, Field(alias="ubiNum")]\n\n\n# 2️⃣ Declare the response model:\nclass NumberToWordsResponse(SuccessfulResponse):\n    __root__: str\n\n\n# 3️⃣ Optionally, declare the error response models:\nclass NumberTooLargeResponse(FaultyResponse):\n    __root__: Literal["number too large"]\n\n\n# 4️⃣ Declare the interface:\nclass SupportsNumberConversion(SupportsService):\n    @operation_name("NumberToWords")\n    def number_to_words(self, request: Body[NumberToWordsRequest]) -> Union[NumberTooLargeResponse, NumberToWordsResponse]:\n        ...\n\n\n# 5️⃣ Bind the service:\nclient = zeep.Client(wsdl="tests/integration/wsdl/NumberConversion.wsdl")\nservice = SupportsNumberConversion.bind(ZeepBackend(client.service))\n\n# 🚀 Call the service:\nresponse = service.number_to_words(NumberToWordsRequest(number=42))\nassert response.unwrap().__root__ == "forty two "\n\n# ☢️ Error classes are automatically derived for error models:\nresponse = service.number_to_words(NumberToWordsRequest(number=-1))\nwith raises(NumberTooLargeResponse.Error):\n    response.raise_for_result()\n```\n',
    'author': 'Pavel Perestoronin',
    'author_email': 'pavel.perestoronin@kpn.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/kpn/combadge',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8.1,<4.0.0',
}


setup(**setup_kwargs)



"""Internal type variables."""

from typing import Callable, TypeVar

from pydantic import BaseModel
from typing_extensions import TypeAlias

RequestT = TypeVar("RequestT", bound=BaseModel)
ResponseT = TypeVar("ResponseT", bound=BaseModel)
ServiceProtocolT = TypeVar("ServiceProtocolT")

_T = TypeVar("_T")
Identity: TypeAlias = Callable[[_T], _T]



from pydantic import BaseModel

from combadge.support.http.abc import RequiresBody
from combadge.support.soap.abc import RequiresOperationName


class Request(RequiresOperationName, RequiresBody, BaseModel):
    """Backend-agnostic SOAP request."""



from typing import Any, Type, Union

from pydantic import BaseModel
from zeep.exceptions import Fault
from zeep.proxy import OperationProxy, ServiceProxy

from combadge.core.binder import BaseBoundService, Signature
from combadge.core.interfaces import SupportsBindServiceMethod, SupportsServiceMethodCall
from combadge.core.request import build_request
from combadge.core.typevars import ResponseT
from combadge.support.soap.request import Request
from combadge.support.soap.response import SoapFaultT
from combadge.support.zeep.backends.base import BaseZeepBackend


class ZeepBackend(BaseZeepBackend[ServiceProxy, OperationProxy], SupportsBindServiceMethod):
    """Synchronous Zeep service."""

    def __call__(
        self,
        request: Request,
        response_type: Type[ResponseT],
        fault_type: Type[SoapFaultT],
    ) -> Union[ResponseT, SoapFaultT]:
        """
        Call the specified service method.

        Args:
            request: prepared request
            response_type: non-fault response model type
            fault_type: SOAP fault model type
        """
        operation = self._get_operation(request.operation_name)
        try:
            response = operation(**request.body.dict(by_alias=True))
        except Fault as e:
            return self._parse_soap_fault(e, fault_type)
        return self._parse_response(response, response_type)

    def bind_method(self, signature: Signature) -> SupportsServiceMethodCall:  # noqa: D102
        response_type, fault_type = self._split_response_type(signature.return_type)

        def resolved_method(service: BaseBoundService, *args: Any, **kwargs: Any) -> BaseModel:
            request = build_request(Request, signature, service, args, kwargs)
            return self(request, response_type, fault_type)

        return resolved_method  # type: ignore[return-value]
